---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--          Copyright (c) 2007, Raul Fechete, Georg Kienesberger         --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Ast2Cfg.Output; use Ast2Cfg.Output;
with Ast2Cfg.Id;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;

package body Ast2Cfg.CFGs is

   overriding
   procedure Init (Cfg:  in out CFG_Object;
                   Oid:  in Id.Pointer;
                   Name: in Unbounded_String;
                   Pred: in Flow_Class_Ptr := null) is -- private
   begin
      Flow_Types.Init(Flow_Object(Cfg), Oid, Name, Pred);
      Cfg.Head := null;
      Cfg.LName := S2u("");
      Declaration_Ptr_List.Init(Cfg.Params);
   end Init;

   not overriding
   procedure Init (Cfg:  in out CFG_Object;
                   Name: in Unbounded_String;
                   Pred: in Flow_Class_Ptr := null) is
   begin
      Init(Cfg, Id.New_Cfg_Id, Name, Pred);
   end Init;

    not overriding
    procedure Free (Cfg: in out CFG_Class_Ptr) is
    begin
       if Cfg /= null then
          Flow_Types.Free(Flow_Class_Ptr(Cfg));
       end if;
    end Free;

   not overriding
   procedure Free_Special (Cfg: in out CFG_Class_Ptr) is
   begin
      if Cfg /= null then
         Declaration_Ptr_List.Free(Cfg.Params);
         Nodes.Free_Recursive(Cfg.Head);
      end if;
   end Free_Special;

   not overriding
   procedure Set_Head (Cfg: in out CFG_Object; Node: in Node_Class_Ptr) is
   begin
      Cfg.Head := Node;
   end Set_Head;

   not overriding
   function Get_Head (Cfg: in CFG_Object) return Node_Class_Ptr is
   begin
      return Cfg.Head;
   end Get_Head;

   not overriding
   procedure Set_LName (Cfg: in out CFG_Object; Name: in Unbounded_String) is
   begin
      Cfg.LName := Name;
   end Set_LName;

   not overriding
   function Get_LName (Cfg: in CFG_Object) return String is
   begin
      return U2s(Cfg.LName);
   end Get_LName;

   not overriding
   procedure Add_Param (Cfg: in out CFG_Object; Param: in Declaration_Ptr) is
   begin
      Declaration_Ptr_List.Append(Cfg.Params, Param);
   end Add_Param;

   not overriding
   function Get_Params (Cfg: in CFG_Object) return Declaration_Ptr_List.Object is
   begin
      return Cfg.Params;
   end Get_Params;

   not overriding
   procedure Increment_Node_Sum (Cfg: in out CFG_Object) is
   begin
      Cfg.Node_Sum := Cfg.Node_Sum + 1;
   end Increment_Node_Sum;

   not overriding
   procedure Decrement_Node_Sum (Cfg: in out CFG_Object) is
   begin
      Cfg.Node_Sum := Cfg.Node_Sum - 1;
   end Decrement_Node_Sum;

   not overriding
   function Get_Node_Sum (Cfg: in CFG_Object) return Long_Long_Natural is
   begin
      return Cfg.Node_Sum;
   end Get_Node_Sum;

end Ast2Cfg.CFGs;
