---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--          Copyright (c) 2007, Raul Fechete, Georg Kienesberger         --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

with Asis; use Asis;
with Asis.Elements; use Asis.Elements;
with Asis.Text; use Asis.Text;

with Ast2Cfg.Output; use Ast2Cfg.Output;
with Ast2Cfg.Utils; use Ast2Cfg.Utils;

package body Ast2Cfg.Impl_Msg is

   procedure Display_Message(Element: Asis.Element) is
   begin
      Display_Message("", Element);
   end Display_Message;

   procedure Display_Message(Text: String; Element: Asis.Element) is
      Element_Kind:       Asis.Element_Kinds;
      Statement_Kind:     Asis.Statement_Kinds;
      Path_Kind:          Asis.Path_Kinds;
      Defining_Name_Kind: Asis.Defining_Name_Kinds;
      Declaration_Kind:   Asis.Declaration_Kinds;
      Pragma_Kind:        Asis.Pragma_Kinds;
      Definition_Kind:    Asis.Definition_Kinds;
      Expression_Kind:    Asis.Expression_Kinds;
      Clause_Kind:        Asis.Clause_Kinds;
      Association_Kind:   Asis.Association_Kinds;
      Elem_Lines:         Asis.Text.Line_List := Lines(Element);
   begin
      Element_Kind := Asis.Elements.Element_Kind (Element);
      case Element_Kind is

         when A_Clause =>
            Clause_Kind := Asis.Elements.Clause_Kind (Element);
            case Clause_Kind is
               when A_Use_Package_Clause =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Use_Package_Clause" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Use_Type_Clause =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Use_Type_Clause" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_With_Clause =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_With_Clause" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when A_Representation_Clause =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Representation_Clause" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Component_Clause =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Component_Clause" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_A_Clause => null;
            end case;

         when An_Expression =>
            Expression_Kind := Asis.Elements.Expression_Kind (Element);
            case Expression_Kind is
               when An_Identifier =>
                  -- Saving each Identifier into the Traversal State Variable
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Identifier" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Selected_Component =>
                  -- SCs are the Name1.Name2.. Addresses
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Selected_Component" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Integer_Literal => -- Immediates
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Integer_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Real_Literal => -- Immediates
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Real_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_String_Literal => -- Immediates
                  Put_Line(Max, Text & "[IMPL:DONE]  A_String_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Character_Literal => -- Immediates
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Character_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Enumeration_Literal => -- Immediates
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Enumeration_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Null_Literal => -- Immediate
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Null_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Parenthesized_Expression => -- Not necessary
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Parenthesized_Expression" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Function_Call =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Function_Call" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Indexed_Component =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Indexed_Component" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Operator_Symbol => -- "or", "and", ..
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Operator_Symbol" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_And_Then_Short_Circuit => -- Implicitely handled
                  Put_Line(Max, Text & "[IMPL:DONE]  An_And_Then_Short_Circuit" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Or_Else_Short_Circuit => -- Implicitely handled
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Or_Else_Short_Circuit" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Allocation_From_Subtype =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Allocation_From_Subtype" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Type_Conversion =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Type_Conversion" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Explicit_Dereference => -- "Var.all" - Implicitely Handled
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Explicit_Dereference" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Attribute_Reference =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Attribute_Reference" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Positional_Array_Aggregate =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Positional_Array_Aggregate" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Record_Aggregate =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Record_Aggregate" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_In_Range_Membership_Test => -- "e.g. Var.all in Type'Class"; handled automatically
                  Put_Line(Max, Text & "[IMPL:DONE]  An_In_Range_Membership_Test" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Slice =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Slice" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when An_In_Type_Membership_Test =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_In_Type_Membership_Test" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Not_In_Type_Membership_Test =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Not_In_Type_Membership_Test" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when An_Extension_Aggregate =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Extension_Aggregate" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Named_Array_Aggregate =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Named_Array_Aggregate" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when A_Not_In_Range_Membership_Test =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Not_In_Range_Membership_Test" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when A_Qualified_Expression =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Qualified_Expression" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Allocation_From_Qualified_Expression =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Allocation_From_Qualified_Expression" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_An_Expression => null;
            end case;

         when A_Path =>
            Path_Kind := Asis.Elements.Path_Kind (Element);
            case Path_Kind is
               -- PATHS ARE THE BEGINNING NODES OF IF/CASE/SELECT BRANCHES
               when An_If_Path =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_If_Path" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Elsif_Path =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Elsif_Path" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Else_Path =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Else_Path" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Select_Path =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Select_Path" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Or_Path =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Or_Path" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Case_Path =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Case_Path" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Then_Abort_Path =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Then_Abort_Path" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_A_Path => null;
            end case;

         when A_Definition =>
            Definition_Kind := Asis.Elements.Definition_Kind (Element);
            case Definition_Kind is
               when A_Protected_Definition => -- Internal container of a Protected Type Declaration
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Protected_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Others_Choice => -- Implicitely handled
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Others_Choice" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Null_Component => -- Irrelevant
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Null_Component" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Null_Record_Definition => -- Irrelevant
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Null_Record_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Type_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Type_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Subtype_Indication =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Subtype_Indication" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Component_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Component_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Record_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Record_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Discrete_Subtype_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Discrete_Subtype_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Discrete_Range => -- e.g. "1..3"
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Discrete_Range" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Task_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Task_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Formal_Type_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Formal_Type_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Constraint => -- e.g. "Type(a..b)", Irrelevant
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Constraint" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Known_Discriminant_Part => -- "Wrapper" for a Discriminant Association
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Known_Discriminant_Part" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Private_Type_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Private_Type_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Tagged_Private_Type_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Tagged_Private_Type_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Private_Extension_Definition =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Private_Extension_Definition" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when An_Unknown_Discriminant_Part =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Unknown_Discriminant_Part" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Variant_Part =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Variant_Part" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Variant =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Variant" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_A_Definition => null;
               when others =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  others reached" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
            end case;

         when A_Pragma =>
            Pragma_Kind := Asis.Elements.Pragma_Kind (Element);
            case Pragma_Kind is
               when An_All_Calls_Remote_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_All_Calls_Remote_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Asynchronous_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Asynchronous_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Atomic_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Atomic_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Atomic_Components_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Atomic_Components_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Attach_Handler_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Attach_Handler_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Controlled_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Controlled_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Convention_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Convention_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Discard_Names_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Discard_Names_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Elaborate_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Elaborate_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Elaborate_All_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Elaborate_All_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Elaborate_Body_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Elaborate_Body_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Export_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Export_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Import_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Import_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Inline_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Inline_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Inspection_Point_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Inspection_Point_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Interrupt_Handler_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Interrupt_Handler_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Interrupt_Priority_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Interrupt_Priority_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Linker_Options_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Linker_Options_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_List_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_List_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Locking_Policy_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Locking_Policy_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Normalize_Scalars_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Normalize_Scalars_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Optimize_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Optimize_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Pack_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Pack_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Page_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Page_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Preelaborate_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Preelaborate_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Priority_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Priority_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Pure_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Pure_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Queuing_Policy_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Queuing_Policy_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Remote_Call_Interface_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Remote_Call_Interface_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Remote_Types_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Remote_Types_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Restrictions_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Restrictions_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Reviewable_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Reviewable_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Shared_Passive_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Shared_Passive_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Storage_Size_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Storage_Size_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Suppress_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Suppress_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Task_Dispatching_Policy_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Task_Dispatching_Policy_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Volatile_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Volatile_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Volatile_Components_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Volatile_Components_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Implementation_Defined_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Implementation_Defined_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Unknown_Pragma =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Unknown_Pragma" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_A_Pragma => null;
               when others =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  others reached" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
            end case;

         when A_Statement =>
            Statement_Kind := Asis.Elements.Statement_Kind (Element);
            case Statement_Kind is
               when A_Null_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Null_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Entry_Call_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Entry_Call_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Accept_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Accept_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Exit_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Exit_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_If_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_If_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Loop_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Loop_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_While_Loop_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_While_Loop_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_For_Loop_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_For_Loop_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Case_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Case_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Procedure_Call_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Procedure_Call_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Return_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Return_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Block_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Block_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Raise_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Raise_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Goto_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Goto_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Requeue_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Requeue_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Requeue_Statement_With_Abort =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Requeue_Statement_With_Abort" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Selective_Accept_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Selective_Accept_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Terminate_Alternative_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Terminate_Alternative_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Timed_Entry_Call_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Timed_Entry_Call_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Conditional_Entry_Call_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Conditional_Entry_Call_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Asynchronous_Select_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Asynchronous_Select_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Abort_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Abort_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Delay_Until_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Delay_Until_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Delay_Relative_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Delay_Relative_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Assignment_Statement =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Assignment_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when A_Code_Statement =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Code_Statement" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_A_Statement => null;
               when others =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  others reached" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
            end case;

         when A_Declaration =>
            Declaration_Kind := Asis.Elements.Declaration_Kind (Element);
            case Declaration_Kind is
               when A_Procedure_Body_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Procedure_Body_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Function_Body_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Function_Body_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Package_Body_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Package_Body_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Entry_Body_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Entry_Body_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Variable_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Variable_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Package_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Package_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Procedure_Declaration => -- Handled by Body
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Procedure_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Function_Declaration => -- Handled by Body
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Function_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Entry_Declaration => -- Handled by Body
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Entry_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Package_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Package_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Procedure_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Procedure_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Function_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Function_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Object_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Object_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Exception_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Exception_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Generic_Package_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Generic_Package_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Generic_Procedure_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Generic_Procedure_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Generic_Function_Renaming_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Generic_Function_Renaming_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Package_Instantiation =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Package_Instantiation" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Procedure_Instantiation =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Procedure_Instantiation" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Function_Instantiation =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Function_Instantiation" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Generic_Procedure_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Generic_Procedure_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Generic_Function_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Generic_Function_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Generic_Package_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Generic_Package_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Formal_Object_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Formal_Object_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Formal_Type_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Formal_Type_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Formal_Procedure_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Formal_Procedure_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Formal_Function_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Formal_Function_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Formal_Package_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Formal_Package_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Formal_Package_Declaration_With_Box =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Formal_Package_Declaration_With_Box" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Task_Type_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Task_Type_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Single_Task_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Single_Task_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Task_Body_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Task_Body_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Protected_Type_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Protected_Type_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Single_Protected_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Single_Protected_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Protected_Body_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Protected_Body_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Component_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Component_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Exception_Declaration => -- Not necessary
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Exception_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Entry_Index_Specification =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Entry_Index_Specification" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Procedure_Body_Stub =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Procedure_Body_Stub" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Function_Body_Stub =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Function_Body_Stub" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Task_Body_Stub =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Task_Body_Stub" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Package_Body_Stub =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Package_Body_Stub" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Protected_Body_Stub =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Protected_Body_Stub" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Ordinary_Type_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Ordinary_Type_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Loop_Parameter_Specification =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Loop_Parameter_Specification" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Enumeration_Literal_Specification => -- Literals are Immediates
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Enumeration_Literal_Specification" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Parameter_Specification =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Parameter_Specification" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Discriminant_Specification =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Discriminant_Specification" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Subtype_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Subtype_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Constant_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Constant_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Deferred_Constant_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Deferred_Constant_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Private_Type_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Private_Type_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Private_Extension_Declaration =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Private_Extension_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when An_Incomplete_Type_Declaration =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Incomplete_Type_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Integer_Number_Declaration =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  An_Integer_Number_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Real_Number_Declaration =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Real_Number_Declaration" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Choice_Parameter_Specification =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Choice_Parameter_Specification" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_A_Declaration => null;
               when others =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  others reached" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
            end case;

         when A_Defining_Name =>
            Defining_Name_Kind := Asis.Elements.Defining_Name_Kind (Element);
            case Defining_Name_Kind is
               -- THE LABELS ARE DEFINING NAMES
               when A_Defining_Identifier =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Defining_Identifier" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Defining_Character_Literal =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Defining_Character_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Defining_Enumeration_Literal =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Defining_Enumeration_Literal" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Defining_Operator_Symbol =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Defining_Operator_Symbol" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Defining_Expanded_Name =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Defining_Expanded_Name" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_A_Defining_Name => null;
            end case;

         when An_Association =>
            Association_Kind := Asis.Elements.Association_Kind (Element);
            case Association_Kind is
               when A_Parameter_Association =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Parameter_Association" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Generic_Association =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Generic_Association" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Discriminant_Association =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Discriminant_Association" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when An_Array_Component_Association =>
                  Put_Line(Max, Text & "[IMPL:DONE]  An_Array_Component_Association" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when A_Record_Component_Association =>
                  Put_Line(Max, Text & "[IMPL:DONE]  A_Record_Component_Association" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

               when A_Pragma_Argument_Association =>
                  Put_Line(Debug, Text & " ** [IMPL:TODO]  A_Pragma_Argument_Association" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));
               when Not_An_Association => null;
            end case;

         when An_Exception_Handler =>
            Put_Line(Max, Text & "[IMPL:DONE]  An_Exception_Handler" & ": " & W2s(Line_Image(Elem_Lines(Elem_Lines'First))));

         when Not_An_Element => null;

      end case;

   end Display_Message;

end Ast2Cfg.Impl_Msg;
