---------------------------------------------------------------------------
--                                                                       --
--  Ast2Cfg is an ASIS application that generates a CFG out of the AST.  --
--                          http://cfg.w3x.org                           --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
--          Copyright (c) 2007, Raul Fechete, Georg Kienesberger         --
--                                                                       --
---------------------------------------------------------------------------
--                                                                       --
-- This file is part of Ast2Cfg.                                         --
--                                                                       --
-- Ast2Cfg is free software; you can redistribute it and/or modify       --
-- it under the terms of the GNU General Public License as published by  --
-- the Free Software Foundation; either version 3 of the License, or     --
-- (at your option) any later version.                                   --
--                                                                       --
-- Ast2Cfg is distributed in the hope that it will be useful,            --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of        --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         --
-- GNU General Public License for more details.                          --
--                                                                       --
-- You should have received a copy of the GNU General Public License     --
-- along with this program.  If not, see <http://www.gnu.org/licenses/>. --
--                                                                       --
---------------------------------------------------------------------------

package body Ast2Cfg.Union_Find is

   -- call Makeset for all elements
   procedure Init_Sets (S: in out Sets) is
   begin
      for I in S'Range loop
         Makeset(S, I);
      end loop;
   end Init_Sets;

   -- creates a set containing only X
   procedure Makeset(S: in out Sets; X: in Element) is
   begin
      if X < First or X > Last then
         raise Out_Of_Range;
      end if;

      S(X) := X;
   end Makeset;

   -- finds the canonical element for the set containing X; uses path compression
   procedure Find (S: in out Sets; X: in Element; Canonical: out Element) is
      Root, Cur, Next: Element;
   begin
      if X < First or X > Last then
         raise Out_Of_Range;
      end if;

      Root := S(X);
      while Root /= S(Root) loop
         Root := S(Root);
      end loop;

      -- path compression
      Next := X;
      while Next /= Root loop
         Cur := Next;
         Next := S(Cur);
         S(Cur) := Root;
      end loop;

      Canonical := Root;
   end Find;

   -- links X to Y, so that find on X and all its descendants returns
   -- the canonical element of the set containing Y
   procedure Link (S: in out Sets; X: in Element; Y: in Element) is
   begin
      if X < First or X > Last or
         Y < First or Y > Last then
         raise Out_Of_Range;
      end if;

      S(X) := Y;
   end Link;

   -- links the canonical elements of the sets containing X or Y, respectively
   procedure Union (S: in out Sets; X: in Element; Y: in Element) is
      Fx, Fy: Element;
   begin
      Find(S, X, Fx);
      Find(S, Y, Fy);
      Link(S, Fx, Fy);
   end Union;

end Ast2Cfg.Union_Find;
